from .version import __version__
from .cache import clear_cache
from .config import configure, load_config
from .mpls import collect, get, use, temp

__all__ = [
    "clear_cache",
    "configure", "load_config",
    "collect", "get", "use", "temp"
]



"""
"""
from urllib.request import urlopen, URLError, HTTPError

from .utils import remove_comments
from .cache import CACHE
from .config import CONFIG, MPLS_TYPES

import json
import matplotlib as mpl
import matplotlib.style
import logging

logger = logging.getLogger(__name__)


def __get(name, type, **kwargs):
    """

    Parameters
    ----------
    name: str

    type: str

    kwargs:
    - style_url: str

    - stylelib_url: str

    - stylelib_format: str

    - ignore_cache: bool

    Raises
    ------
    IOError:
    HTTPError:
    URLError:
    JSONDecodeError:
    """
    folder    = kwargs.get('stylelib_url', CONFIG['stylelib_url'])
    file_path = kwargs.get('stylelib_format', CONFIG['stylelib_format'])
    style_url = kwargs.get('style_url', folder+file_path).format(type=type, name=name)

    if not kwargs.get('ignore_cache', False) and CACHE.is_cached(type, name):
        logger.debug('loading {} file from cache'.format(type))
        with open(CACHE.file_path(type=type, name=name), 'r') as f:
            content = remove_comments(f.read())
    else:
        try:
            logger.debug('trying to urlopen file: {}'.format(style_url))
            with urlopen(style_url) as f:
                # get file content from specified url
                content = remove_comments(f.read().decode())
            logger.debug('loaded raw {} file from URL'.format(type))
            CACHE.add(type, name, content)
        except ValueError as e:  # style_url is not a valid url
            logger.debug('urlopen failed: {}'.format(str(e)))
            logger.debug('trying to (regular) open file')
            try:
                with open(style_url) as f:
                    # get file content from file path instead
                    content = remove_comments(f.read())
                logger.debug('loaded file from local disk'.format(type))
            except IOError:
                raise
        except HTTPError as e:
            raise
        except URLError as e:
            logger.debug('urlopen failed: {}'.format(str(e)))
            raise

    try:
        logger.debug('converting file content to Python dict')
        # convert file content to python dict
        return json.loads(content)
    except json.JSONDecodeError as e:
        logger.debug('json.loads failed: {}'.format(str(e)))
        raise


def get(name, type, **kwargs):
    """Returns the rcParams specified in the style file given by `name` and `type`.

    Parameters
    ----------
    name: str
        The name of the style.
    type: str
        Any of ('context', 'style', 'palette').
    kwargs:
    - stylelib_url: str
        Overwrite the value in the local config with the specified url.
    - ignore_cache: bool
        Ignore files in the cache and force loading from the stylelib.

    Raises
    ------
    ValueError:
        If `type` is not any of ('context', 'style', 'palette')

    Returns
    -------
    rcParams: dict
        The parameter dict of the file.
    """
    type = str(type)

    params = {}
    if type in MPLS_TYPES:
        params.update(__get(name, type, **kwargs))
    else:
        raise ValueError('unexpected type: {}! Must be any of {!r}'.format(type, MPLS_typeS))

    # color palette hack
    if params.get('axes.prop_cycle'):
        params['axes.prop_cycle'] = mpl.rcsetup.cycler('color', params['axes.prop_cycle'])

    return params


def collect(context=None, style=None, palette=None, **kwargs):
    """Returns the merged rcParams dict of the specified context, style, and palette.

    Parameters
    ----------
    context: str

    style: str

    palette: str

    kwargs:
    -

    Returns
    -------
    rcParams: dict
        The merged parameter dicts of the specified context, style, and palette.

    Notes
    -----
    The rcParams dicts are loaded and updated in the order: context, style, palette. That means if
    a context parameter is also defined in the style or palette dict, it will be overwritten. There
    is currently no checking being done to avoid this.
    """
    params = {}
    if context:
        params.update(get(context, 'context', **kwargs))
    if style:
        params.update(get(style, 'style', **kwargs))
    if palette:
        params.update(get(palette, 'palette', **kwargs))
    return params


def use(*args, context=None, style=None, palette=None, **kwargs):
    """

    Parameters
    ----------
    args:

    context: str or None

    style: str or None

    palette: str or None

    kwargs:
    - reset

    Raises
    ------
    ValueError:

    """
    if kwargs.get('reset', False):
        styles = ['default', ]
    else:
        styles = []

    styles.extend(list(args))
    styles.append(collect(context=context, style=style, palette=palette, **kwargs))
    # apply mpls styles
    return mpl.style.use(styles)


def temp(*args, context=None, style=None, palette=None, **kwargs):
    """

    Parameters
    ----------
    args:

    context: str or None

    style: str or None

    palette: str or None

    kwargs:
    - reset

    Raises
    ------
    ValueError:

    """
    # apply specified matplotlib styles and reset if specified
    styles = list(args)
    styles.append(collect(context=context, style=style, palette=palette, **kwargs))
    return mpl.style.context(styles, after_reset=kwargs.get('reset', False))



"""Utility functions

- remove_comments(text)
"""
import re


# Credits to:
# https://stackoverflow.com/questions/2319019/using-regex-to-remove-comments-from-source-files
# first group captures quoted strings (double or single)
# second group captures comments (//single-line or /* multi-line */)
COMMENT_PATTERN = re.compile("(\".*?\"|\'.*?\')|(/\*.*?\*/|//[^\r\n]*$)", re.MULTILINE | re.DOTALL)


def remove_comments(text):
    """Remove C and Javascript style comments like // [...] and /* [...] */ from
    the given text.
    """
    def _replacer(match):
        # if the 2nd group (capturing comments) is not None,
        # it means we have captured a non-quoted (real) comment string.
        if match.group(2) is not None:
            return ""  # so we will return empty to remove the comment
        else:  # otherwise, we will return the 1st group
            return match.group(1)  # captured quoted-string

    return COMMENT_PATTERN.sub(_replacer, text)



from .config import CACHE_DIR, CONFIG

import os
import shutil
import logging
import hashlib

logger = logging.getLogger(__name__)


class Cache(object):

    def __init__(self, cache_dir):
        self.cache_dir = cache_dir

    def __del__(self):
        if CONFIG['auto_clear_cache']:
            self.clear()

    def file_path(self, type, name):
        return os.path.join(self.host_folder(CONFIG['stylelib_url']), CONFIG['stylelib_format'].format(type=type, name=name))

    def host_folder(self, host_url):
        return os.path.join(self.cache_dir, hashlib.sha1(host_url.encode()).hexdigest())

    def is_cached(self, type, name):
        return os.path.exists(self.file_path(type, name))

    def add(self, type, name, content):
        if self.is_cached(type, name):
            return

        path = self.file_path(type, name)
        dir_path = os.path.dirname(path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        try:
            with open(path, 'w') as cache_file:
                cache_file.write(content)
        except OSError:
            pass

        logger.debug('added {} file "{}" to cache'.format(type, name))

    def clear(self, host_url=None):
        if not host_url:
            clean_up_dir = self.cache_dir
        else:
            clean_up_dir = self.host_folder(host_url)

        logger.debug('cleaning up cache for stylelib host: {}'.format(host_url))
        if os.path.exists(clean_up_dir):
            shutil.rmtree(clean_up_dir, ignore_errors=True)


CACHE = Cache(CACHE_DIR)


def clear_cache(host_url=None):
    """Clears the file cache of the specified host or the complete cache if host_url is
    not specified.

    Can be useful if the original style file has changed but an old version is still
    available in the cache. Alternatively you can call any of the frontend methods
    (get, use, temp) with the optional `ignore_cache` parameter.
    """
    CACHE.clear(host_url)



import os
import json
import logging

logger = logging.getLogger(__name__)

MPLS_TYPES = ('context', 'style', 'palette')

CONFIG_DIR = os.path.expanduser('~/.mpls')
CONFIG_FILE = os.path.join(CONFIG_DIR, 'config.json')
CONFIG = {
    'stylelib_url': "https://raw.githubusercontent.com/scruffy-t/mpls/master/stylelib/",
    'stylelib_format': "{type}/{name}.json",
    'auto_clear_cache': False,
    'enable_logging': False
}

CACHE_DIR = os.path.join(CONFIG_DIR, 'cache')


def load_config():
    global CONFIG
    try:
        CONFIG = json.load(open(CONFIG_FILE, 'r'))
        logger.debug('using config file {}'.format(CONFIG_FILE))
    except OSError:
        logger.debug('no config file found, using default config')

    if CONFIG['enable_logging']:
        logging.basicConfig(level=logging.DEBUG)

load_config()


def configure(save=False, **kwargs):
    """

    Parameters
    ----------
    save: bool
        save config to file after it was updated
    kwargs:
    - stylelib_url: str

    - stylelib_format: str

    - auto_clear_cache: bool

    - enable_logging: bool

    """
    if 'stylelib_url' in kwargs:
        # TODO: is it safe to do that with a web URL?
        kwargs['stylelib_url'] = os.path.expanduser(kwargs['stylelib_url'])

    CONFIG.update(kwargs)

    if save:
        if not os.path.exists(CONFIG_DIR):
            os.mkdir(CONFIG_DIR)
        json.dump(CONFIG, open(CONFIG_FILE, 'w'))
        logger.debug('saved customized config to {}'.format(CONFIG_FILE))



from setuptools import setup
from codecs import open
from os import path

from mpls import __version__

here = path.abspath(path.dirname(__file__))


def readme():
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        return f.read()


setup(
    name='mpls',
    version=__version__,
    description='An open library of matplotlib styles',
    long_description=readme(),

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],

    keywords='plotting matplotlib styles',
    url='https://github.com/scruffy-t/mpls',
    download_url='https://github.com/scruffy-t/mpls/archive/{}.tar.gz'.format(__version__),
    author='Tobias Schruff',
    author_email='tobias.schruff@gmail.com',
    license='BSD (3-clause)',
    packages=['mpls'],
    install_requires=['matplotlib'],
    zip_safe=False
)
